@@ -1,6 +1,6 @@
 /* dirmngr.c - Keyserver and X.509 LDAP access
  * Copyright (C) 2002 Klarälvdalens Datakonsult AB
- * Copyright (C) 2003, 2004, 2006, 2007, 2008, 2010, 2011, 2020 g10 Code GmbH
+ * Copyright (C) 2003-2004, 2006-2007, 2008, 2010-2011, 2020-2021 g10 Code GmbH
  * Copyright (C) 2014 Werner Koch
  *
  * This file is part of GnuPG.
@@ -107,6 +107,7 @@ enum cmd_and_opt_values {
   oDebugWait,
   oDebugLevel,
   oGnutlsDebug,
+  oDebugCacheExpiredCerts,
   oNoGreeting,
   oNoOptions,
   oHomedir,
@@ -145,6 +146,7 @@ enum cmd_and_opt_values {
   oHTTPWrapperProgram,
   oIgnoreCert,
   oIgnoreCertExtension,
+  oIgnoreCRLExtension,
   oUseTor,
   oNoUseTor,
   oKeyServer,
@@ -156,6 +158,7 @@ enum cmd_and_opt_values {
   oConnectTimeout,
   oConnectQuickTimeout,
   oListenBacklog,
+  oCompatibilityFlags,
   aTest
 };
 
@@ -163,6 +166,10 @@ enum cmd_and_opt_values {
 
 static ARGPARSE_OPTS opts[] = {
 
+  ARGPARSE_c (aGPGConfList, "gpgconf-list", "@"),
+  ARGPARSE_c (aGPGConfTest, "gpgconf-test", "@"),
+  ARGPARSE_c (aGPGConfVersions, "gpgconf-versions", "@"),
+
   ARGPARSE_group (300, N_("@Commands:\n ")),
 
   ARGPARSE_c (aServer,   "server",  N_("run in server mode (foreground)") ),
@@ -171,48 +178,97 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_c (aSupervised,  "supervised", N_("run in supervised mode")),
 #endif
   ARGPARSE_c (aListCRLs, "list-crls", N_("list the contents of the CRL cache")),
-  ARGPARSE_c (aLoadCRL,  "load-crl",  N_("|FILE|load CRL from FILE into cache")),
+  ARGPARSE_c (aLoadCRL,  "load-crl", N_("|FILE|load CRL from FILE into cache")),
   ARGPARSE_c (aFetchCRL, "fetch-crl", N_("|URL|fetch a CRL from URL")),
   ARGPARSE_c (aShutdown, "shutdown",  N_("shutdown the dirmngr")),
   ARGPARSE_c (aFlush,    "flush",     N_("flush the cache")),
-  ARGPARSE_c (aGPGConfList, "gpgconf-list", "@"),
-  ARGPARSE_c (aGPGConfTest, "gpgconf-test", "@"),
-  ARGPARSE_c (aGPGConfVersions, "gpgconf-versions", "@"),
 
-  ARGPARSE_group (301, N_("@\nOptions:\n ")),
+
+  ARGPARSE_header (NULL, N_("Options used for startup")),
+
+  ARGPARSE_s_n (oNoDetach, "no-detach", N_("do not detach from the console")),
+  ARGPARSE_s_n (oSh,       "sh",        N_("sh-style command output")),
+  ARGPARSE_s_n (oCsh,      "csh",       N_("csh-style command output")),
+  ARGPARSE_s_n (oStealSocket, "steal-socket", "@"),
+  ARGPARSE_s_s (oHomedir, "homedir", "@"),
+  ARGPARSE_conffile (oOptions,  "options", N_("|FILE|read options from FILE")),
+  ARGPARSE_noconffile (oNoOptions, "no-options", "@"),
+
+
+  ARGPARSE_header ("Monitor", N_("Options controlling the diagnostic output")),
 
   ARGPARSE_s_n (oVerbose,  "verbose",   N_("verbose")),
   ARGPARSE_s_n (oQuiet,    "quiet",     N_("be somewhat more quiet")),
-  ARGPARSE_s_n (oSh,       "sh",        N_("sh-style command output")),
-  ARGPARSE_s_n (oCsh,      "csh",       N_("csh-style command output")),
-  ARGPARSE_conffile (oOptions, "options", N_("|FILE|read options from FILE")),
+  ARGPARSE_s_n (oNoGreeting, "no-greeting", "@"),
   ARGPARSE_s_s (oDebugLevel, "debug-level",
                 N_("|LEVEL|set the debugging level to LEVEL")),
-  ARGPARSE_s_n (oNoDetach, "no-detach", N_("do not detach from the console")),
-  ARGPARSE_s_n (oStealSocket, "steal-socket", "@"),
+  ARGPARSE_s_s (oDebug,    "debug", "@"),
+  ARGPARSE_s_n (oDebugAll, "debug-all", "@"),
+  ARGPARSE_s_i (oGnutlsDebug, "gnutls-debug", "@"),
+  ARGPARSE_s_i (oGnutlsDebug, "tls-debug", "@"),
+  ARGPARSE_s_i (oDebugWait, "debug-wait", "@"),
   ARGPARSE_s_s (oLogFile,  "log-file",
                 N_("|FILE|write server mode logs to FILE")),
-  ARGPARSE_s_n (oBatch,    "batch",       N_("run without asking a user")),
-  ARGPARSE_s_n (oForce,    "force",       N_("force loading of outdated CRLs")),
-  ARGPARSE_s_n (oAllowOCSP, "allow-ocsp", N_("allow sending OCSP requests")),
+
+
+  ARGPARSE_header ("Configuration",
+                   N_("Options controlling the configuration")),
+
   ARGPARSE_s_n (oAllowVersionCheck, "allow-version-check",
                 N_("allow online software version check")),
+  ARGPARSE_s_i (oListenBacklog, "listen-backlog", "@"),
+  ARGPARSE_s_i (oMaxReplies, "max-replies",
+                N_("|N|do not return more than N items in one query")),
+  ARGPARSE_s_u (oFakedSystemTime, "faked-system-time", "@"), /*(epoch time)*/
+  ARGPARSE_s_n (oDisableCheckOwnSocket, "disable-check-own-socket", "@"),
+  ARGPARSE_s_s (oIgnoreCert,"ignore-cert", "@"),
+  ARGPARSE_s_s (oIgnoreCertExtension,"ignore-cert-extension", "@"),
+  ARGPARSE_s_s (oIgnoreCRLExtension,"ignore-crl-extension", "@"),
+
+
+  ARGPARSE_header ("Network", N_("Network related options")),
+
+  ARGPARSE_s_n (oUseTor, "use-tor", N_("route all network traffic via Tor")),
+  ARGPARSE_s_n (oNoUseTor, "no-use-tor", "@"),
+  ARGPARSE_s_n (oDisableIPv4, "disable-ipv4", "@"),
+  ARGPARSE_s_n (oDisableIPv6, "disable-ipv6", "@"),
+  ARGPARSE_s_n (oStandardResolver, "standard-resolver", "@"),
+  ARGPARSE_s_n (oRecursiveResolver, "recursive-resolver", "@"),
+  ARGPARSE_s_i (oResolverTimeout, "resolver-timeout", "@"),
+  ARGPARSE_s_s (oNameServer, "nameserver", "@"),
+  ARGPARSE_s_i (oConnectTimeout, "connect-timeout", "@"),
+  ARGPARSE_s_i (oConnectQuickTimeout, "connect-quick-timeout", "@"),
+
+
+  ARGPARSE_header ("Keyserver", N_("Configuration for Keyservers")),
+
+  ARGPARSE_s_s (oKeyServer, "keyserver",
+                N_("|URL|use keyserver at URL")),
+  ARGPARSE_s_s (oHkpCaCert, "hkp-cacert",
+                N_("|FILE|use the CA certificates in FILE for HKP over TLS")),
+
+
+  ARGPARSE_header ("HTTP", N_("Configuration for HTTP servers")),
+
   ARGPARSE_s_n (oDisableHTTP, "disable-http", N_("inhibit the use of HTTP")),
-  ARGPARSE_s_n (oDisableLDAP, "disable-ldap", N_("inhibit the use of LDAP")),
   ARGPARSE_s_n (oIgnoreHTTPDP,"ignore-http-dp",
                 N_("ignore HTTP CRL distribution points")),
-  ARGPARSE_s_n (oIgnoreLDAPDP,"ignore-ldap-dp",
-                N_("ignore LDAP CRL distribution points")),
-  ARGPARSE_s_n (oIgnoreOCSPSvcUrl, "ignore-ocsp-service-url",
-                N_("ignore certificate contained OCSP service URLs")),
-
   ARGPARSE_s_s (oHTTPProxy,  "http-proxy",
                 N_("|URL|redirect all HTTP requests to URL")),
+  ARGPARSE_s_n (oHonorHTTPProxy, "honor-http-proxy",
+                N_("use system's HTTP proxy setting")),
+  ARGPARSE_s_s (oLDAPWrapperProgram, "ldap-wrapper-program", "@"),
+
+
+  ARGPARSE_header ("LDAP", N_("Configuration of LDAP servers to use")),
+
+  ARGPARSE_s_n (oDisableLDAP, "disable-ldap", N_("inhibit the use of LDAP")),
+  ARGPARSE_s_n (oIgnoreLDAPDP,"ignore-ldap-dp",
+                N_("ignore LDAP CRL distribution points")),
   ARGPARSE_s_s (oLDAPProxy,  "ldap-proxy",
                 N_("|HOST|use HOST for LDAP queries")),
   ARGPARSE_s_n (oOnlyLDAPProxy, "only-ldap-proxy",
                 N_("do not use fallback hosts with --ldap-proxy")),
-
   ARGPARSE_s_s (oLDAPServer, "ldapserver",
                 N_("|SPEC|use this keyserver to lookup keys")),
   ARGPARSE_s_s (oLDAPFile, "ldapserverlist-file",
@@ -223,6 +279,12 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_i (oLDAPTimeout, "ldaptimeout",
                 N_("|N|set LDAP timeout to N seconds")),
 
+
+  ARGPARSE_header ("OCSP", N_("Configuration for OCSP")),
+
+  ARGPARSE_s_n (oAllowOCSP, "allow-ocsp", N_("allow sending OCSP requests")),
+  ARGPARSE_s_n (oIgnoreOCSPSvcUrl, "ignore-ocsp-service-url",
+                N_("ignore certificate contained OCSP service URLs")),
   ARGPARSE_s_s (oOCSPResponder, "ocsp-responder",
                 N_("|URL|use OCSP responder at URL")),
   ARGPARSE_s_s (oOCSPSigner, "ocsp-signer",
@@ -231,43 +293,21 @@ static ARGPARSE_OPTS opts[] = {
   ARGPARSE_s_i (oOCSPMaxPeriod,    "ocsp-max-period", "@"),
   ARGPARSE_s_i (oOCSPCurrentPeriod, "ocsp-current-period", "@"),
 
-  ARGPARSE_s_i (oMaxReplies, "max-replies",
-                N_("|N|do not return more than N items in one query")),
 
-  ARGPARSE_s_s (oNameServer, "nameserver", "@"),
-  ARGPARSE_s_s (oKeyServer, "keyserver", "@"),
-  ARGPARSE_s_s (oHkpCaCert, "hkp-cacert",
-                N_("|FILE|use the CA certificates in FILE for HKP over TLS")),
+  ARGPARSE_header (NULL, N_("Other options")),
 
-  ARGPARSE_s_n (oUseTor, "use-tor", N_("route all network traffic via Tor")),
-  ARGPARSE_s_n (oNoUseTor, "no-use-tor", "@"),
-
-  ARGPARSE_s_n (oDisableIPv4, "disable-ipv4", "@"),
-  ARGPARSE_s_n (oDisableIPv6, "disable-ipv6", "@"),
+  ARGPARSE_s_n (oForce,    "force",    N_("force loading of outdated CRLs")),
 
   ARGPARSE_s_s (oSocketName, "socket-name", "@"),  /* Only for debugging.  */
+  ARGPARSE_s_n (oDebugCacheExpiredCerts, "debug-cache-expired-certs", "@"),
+  ARGPARSE_s_s (oCompatibilityFlags, "compatibility-flags", "@"),
 
-  ARGPARSE_s_u (oFakedSystemTime, "faked-system-time", "@"), /*(epoch time)*/
-  ARGPARSE_s_s (oDebug,    "debug", "@"),
-  ARGPARSE_s_n (oDebugAll, "debug-all", "@"),
-  ARGPARSE_s_i (oGnutlsDebug, "gnutls-debug", "@"),
-  ARGPARSE_s_i (oGnutlsDebug, "tls-debug", "@"),
-  ARGPARSE_s_i (oDebugWait, "debug-wait", "@"),
-  ARGPARSE_s_n (oDisableCheckOwnSocket, "disable-check-own-socket", "@"),
-  ARGPARSE_s_n (oNoGreeting, "no-greeting", "@"),
-  ARGPARSE_s_s (oHomedir, "homedir", "@"),
-  ARGPARSE_s_s (oLDAPWrapperProgram, "ldap-wrapper-program", "@"),
+  ARGPARSE_header (NULL, ""),  /* Stop the header group.  */
+
+  /* Not yet used options.  */
+  ARGPARSE_s_n (oBatch,    "batch",       "@"),
   ARGPARSE_s_s (oHTTPWrapperProgram, "http-wrapper-program", "@"),
-  ARGPARSE_s_n (oHonorHTTPProxy, "honor-http-proxy", "@"),
-  ARGPARSE_s_s (oIgnoreCertExtension,"ignore-cert-extension", "@"),
-  ARGPARSE_s_s (oIgnoreCert,"ignore-cert", "@"),
-  ARGPARSE_s_n (oStandardResolver, "standard-resolver", "@"),
-  ARGPARSE_s_n (oRecursiveResolver, "recursive-resolver", "@"),
-  ARGPARSE_s_i (oResolverTimeout, "resolver-timeout", "@"),
-  ARGPARSE_s_i (oConnectTimeout, "connect-timeout", "@"),
-  ARGPARSE_s_i (oConnectQuickTimeout, "connect-quick-timeout", "@"),
-  ARGPARSE_s_i (oListenBacklog, "listen-backlog", "@"),
-  ARGPARSE_noconffile (oNoOptions, "no-options", "@"),
+
 
   ARGPARSE_group (302,N_("@\n(See the \"info\" manual for a complete listing "
                          "of all commands and options)\n")),
@@ -292,6 +332,14 @@ static struct debug_flags_s debug_flags [] =
     { 77, NULL } /* 77 := Do not exit on "help" or "?".  */
   };
 
+/* The list of compatibility flags.  */
+static struct compatibility_flags_s compatibility_flags [] =
+  {
+    { COMPAT_OCSP_SHA256_CERTID, "ocsp-sha256-certid" },
+    { 0, NULL }
+  };
+
+
 #define DEFAULT_MAX_REPLIES 10
 #define DEFAULT_LDAP_TIMEOUT 15  /* seconds */
 
@@ -435,9 +483,9 @@ my_strusage( int level )
 
 
 /* Callback from libksba to hash a provided buffer.  Our current
-   implementation does only allow SHA-1 for hashing. This may be
-   extended by mapping the name, testing for algorithm availibility
-   and adjust the length checks accordingly. */
+ * implementation does only allow SHA-1 and SHA-256 for hashing. This
+ * may be extended by mapping the name, testing for algorithm
+ * availibility and adjust the length checks accordingly.  */
 static gpg_error_t
 my_ksba_hash_buffer (void *arg, const char *oid,
                      const void *buffer, size_t length, size_t resultsize,
@@ -445,12 +493,22 @@ my_ksba_hash_buffer (void *arg, const char *oid,
 {
   (void)arg;
 
-  if (oid && strcmp (oid, "1.3.14.3.2.26"))
+  if (!oid || !strcmp (oid, "1.3.14.3.2.26"))
+    {
+      if (resultsize < 20)
+        return gpg_error (GPG_ERR_BUFFER_TOO_SHORT);
+      gcry_md_hash_buffer (GCRY_MD_SHA1, result, buffer, length);
+      *resultlen = 20;
+    }
+  else if (!strcmp (oid, "2.16.840.1.101.3.4.2.1"))
+    {
+      if (resultsize < 32)
+        return gpg_error (GPG_ERR_BUFFER_TOO_SHORT);
+      gcry_md_hash_buffer (GCRY_MD_SHA256, result, buffer, length);
+      *resultlen = 32;
+    }
+  else
     return gpg_error (GPG_ERR_NOT_SUPPORTED);
-  if (resultsize < 20)
-    return gpg_error (GPG_ERR_BUFFER_TOO_SHORT);
-  gcry_md_hash_buffer (2, result, buffer, length);
-  *resultlen = 20;
   return 0;
 }
 
@@ -587,6 +645,15 @@ dirmngr_use_tor (void)
 }
 
 
+/* This is somewhat similar to dirmngr_use_tor but avoids a trial
+ * connect and may thus be faster for this special case.  */
+int
+dirmngr_never_use_tor_p (void)
+{
+  return tor_mode == TOR_MODE_NEVER;
+}
+
+
 static void
 wrong_args (const char *text)
 {
@@ -650,6 +717,7 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
           opt.ignored_certs = tmp;
         }
       FREE_STRLIST (opt.ignored_cert_extensions);
+      FREE_STRLIST (opt.ignored_crl_extensions);
       http_register_tls_ca (NULL);
       FREE_STRLIST (hkp_cacert_filenames);
       FREE_STRLIST (opt.keyserver);
@@ -663,6 +731,8 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
       opt.connect_quick_timeout = 0;
       opt.ldaptimeout = DEFAULT_LDAP_TIMEOUT;
       ldapserver_list_needs_reset = 1;
+      opt.debug_cache_expired_certs = 0;
+      opt.compat_flags = 0;
       return 1;
     }
 
@@ -759,6 +829,10 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
       add_to_strlist (&opt.ignored_cert_extensions, pargs->r.ret_str);
       break;
 
+    case oIgnoreCRLExtension:
+      add_to_strlist (&opt.ignored_crl_extensions, pargs->r.ret_str);
+      break;
+
     case oUseTor:
       tor_mode = TOR_MODE_FORCE;
       break;
@@ -821,6 +895,19 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
       opt.ldaptimeout = pargs->r.ret_int;
       break;
 
+    case oDebugCacheExpiredCerts:
+      opt.debug_cache_expired_certs = 0;
+      break;
+
+    case oCompatibilityFlags:
+      if (parse_compatibility_flags (pargs->r.ret_str, &opt.compat_flags,
+                                     compatibility_flags))
+        {
+          pargs->r_opt = ARGPARSE_INVALID_ARG;
+          pargs->err = ARGPARSE_PRINT_WARNING;
+        }
+      break;
+
     default:
       return 0; /* Not handled. */
     }
@@ -837,7 +924,7 @@ parse_rereadable_options (ARGPARSE_ARGS *pargs, int reread)
 /* This fucntion is called after option parsing to adjust some values
  * and call option setup functions.  */
 static void
-post_option_parsing (void)
+post_option_parsing (enum cmd_and_opt_values cmd)
 {
   /* It would be too surpirsing if the quick timeout is larger than
    * the standard value.  */
@@ -845,7 +932,18 @@ post_option_parsing (void)
     opt.connect_quick_timeout = opt.connect_timeout;
 
   set_debug ();
-  set_tor_mode ();
+  /* For certain commands we do not want to set/test for Tor mode
+   * because that is somewhat expensive.  */
+  switch (cmd)
+    {
+    case aGPGConfList:
+    case aGPGConfTest:
+    case aGPGConfVersions:
+      break;
+    default:
+      set_tor_mode ();
+      break;
+    }
 }
 
 
@@ -1158,7 +1256,7 @@ main (int argc, char **argv)
       log_printf ("\n");
     }
 
-  post_option_parsing ();
+  post_option_parsing (cmd);
 
   /* Get LDAP server list from file unless --ldapserver has been used.  */
 #if USE_LDAP
@@ -1544,68 +1642,19 @@ main (int argc, char **argv)
   else if (cmd == aGPGConfList)
     {
       unsigned long flags = 0;
-      char *filename;
       char *filename_esc;
 
-      filename = percent_escape (opt.config_filename, NULL);
-      es_printf ("gpgconf-dirmngr.conf:%lu:\"%s\n",
-              GC_OPT_FLAG_DEFAULT, filename);
-      xfree (filename);
-
-      es_printf ("verbose:%lu:\n", flags | GC_OPT_FLAG_NONE);
-      es_printf ("quiet:%lu:\n", flags | GC_OPT_FLAG_NONE);
       es_printf ("debug-level:%lu:\"none\n", flags | GC_OPT_FLAG_DEFAULT);
-      es_printf ("log-file:%lu:\n", flags | GC_OPT_FLAG_NONE);
-      es_printf ("force:%lu:\n", flags | GC_OPT_FLAG_NONE);
-
-      /* --csh and --sh are mutually exclusive, something we can not
-         express in GPG Conf.  --options is only usable from the
-         command line, really.  --debug-all interacts with --debug,
-         and having both of them is thus problematic.  --no-detach is
-         also only usable on the command line.  --batch is unused.  */
-
-      filename = make_filename (gnupg_homedir (),
-                                "dirmngr_ldapservers.conf",
-                                NULL);
-      filename_esc = percent_escape (filename, NULL);
-      es_printf ("ldapserverlist-file:%lu:\"%s\n", flags | GC_OPT_FLAG_DEFAULT,
-	      filename_esc);
-      xfree (filename_esc);
-      xfree (filename);
-
       es_printf ("ldaptimeout:%lu:%u\n",
-              flags | GC_OPT_FLAG_DEFAULT, DEFAULT_LDAP_TIMEOUT);
+                 flags | GC_OPT_FLAG_DEFAULT, DEFAULT_LDAP_TIMEOUT);
       es_printf ("max-replies:%lu:%u\n",
-              flags | GC_OPT_FLAG_DEFAULT, DEFAULT_MAX_REPLIES);
-      es_printf ("allow-ocsp:%lu:\n", flags | GC_OPT_FLAG_NONE);
-      es_printf ("allow-version-check:%lu:\n", flags | GC_OPT_FLAG_NONE);
-      es_printf ("ocsp-responder:%lu:\n", flags | GC_OPT_FLAG_NONE);
-      es_printf ("ocsp-signer:%lu:\n", flags | GC_OPT_FLAG_NONE);
-
-      es_printf ("faked-system-time:%lu:\n", flags | GC_OPT_FLAG_NONE);
-      es_printf ("no-greeting:%lu:\n", flags | GC_OPT_FLAG_NONE);
-
-      es_printf ("disable-http:%lu:\n", flags | GC_OPT_FLAG_NONE);
-      es_printf ("disable-ldap:%lu:\n", flags | GC_OPT_FLAG_NONE);
-      es_printf ("honor-http-proxy:%lu\n", flags | GC_OPT_FLAG_NONE);
-      es_printf ("http-proxy:%lu:\n", flags | GC_OPT_FLAG_NONE);
-      es_printf ("ldap-proxy:%lu:\n", flags | GC_OPT_FLAG_NONE);
-      es_printf ("only-ldap-proxy:%lu:\n", flags | GC_OPT_FLAG_NONE);
-      es_printf ("ignore-ldap-dp:%lu:\n", flags | GC_OPT_FLAG_NONE);
-      es_printf ("ignore-http-dp:%lu:\n", flags | GC_OPT_FLAG_NONE);
-      es_printf ("ignore-ocsp-service-url:%lu:\n", flags | GC_OPT_FLAG_NONE);
-      /* Note: The next one is to fix a typo in gpgconf - should be
-         removed eventually. */
-      es_printf ("ignore-ocsp-servic-url:%lu:\n", flags | GC_OPT_FLAG_NONE);
-
-      es_printf ("use-tor:%lu:\n", flags | GC_OPT_FLAG_NONE);
+                 flags | GC_OPT_FLAG_DEFAULT, DEFAULT_MAX_REPLIES);
 
       filename_esc = percent_escape (get_default_keyserver (0), NULL);
       es_printf ("keyserver:%lu:\"%s:\n", flags | GC_OPT_FLAG_DEFAULT,
                  filename_esc);
       xfree (filename_esc);
 
-      es_printf ("nameserver:%lu:\n", flags | GC_OPT_FLAG_NONE);
       es_printf ("resolver-timeout:%lu:%u\n",
                  flags | GC_OPT_FLAG_DEFAULT, 0);
     }
@@ -1668,6 +1717,8 @@ dirmngr_deinit_default_ctrl (ctrl_t ctrl)
 
   xfree (ctrl->http_proxy);
   ctrl->http_proxy = NULL;
+  nvc_release (ctrl->rootdse);
+  ctrl->rootdse = NULL;
 }
 
 
@@ -1943,7 +1994,7 @@ reread_configuration (void)
     }
   gnupg_argparse (NULL, &pargs, NULL);  /* Release internal state.  */
   xfree (twopart);
-  post_option_parsing ();
+  post_option_parsing (0);
 }
 
 
@@ -1959,6 +2010,7 @@ dirmngr_sighup_action (void)
   crl_cache_deinit ();
   cert_cache_init (hkp_cacert_filenames);
   crl_cache_init ();
+  http_reinitialize ();
   reload_dns_stuff (0);
   ks_hkp_reload ();
 }
@@ -2474,7 +2526,10 @@ gpgconf_versions (void)
   const char *s;
   int n;
 
-  /* Unfortunately Npth has no way to get the version.  */
+#if NPTH_VERSION_NUMBER >= 0x010800
+  es_fprintf (es_stdout, "* nPth %s (%s)\n\n",
+              npth_get_version (NULL), npth_get_version ("\x01\x02"));
+#endif /*NPTH_VERSION_NUMBER*/
 
   s = get_revision_from_blurb (assuan_check_version ("\x01\x01"), &n);
   es_fprintf (es_stdout, "* Libassuan %s (%.*s)\n\n",
