@@ -54,7 +54,8 @@
 #define SERVERINFO_PGPKEYV2 2 /* Needs "pgpKeyV2" instead of "pgpKey"*/
 #define SERVERINFO_SCHEMAV2 4 /* Version 2 of the Schema.            */
 #define SERVERINFO_NTDS     8 /* Server is an Active Directory.      */
-#define SERVERINFO_GENERIC 16 /* Connected in genric mode.           */
+#define SERVERINFO_GENERIC 16 /* Connected in generic mode.          */
+#define SERVERINFO_CNFPR   32 /* Server uses CN=fingerprint.         */
 
 
 /* The page size requested from the server.  */
@@ -589,7 +590,10 @@ interrogate_ldap_dn (LDAP *ldap_conn, const char *basedn_search,
                 *r_serverinfo |= SERVERINFO_SCHEMAV2;
               if (nfields > 1
                   && !ascii_strcasecmp (fields[1], "ntds"))
-                *r_serverinfo |= SERVERINFO_NTDS;
+                *r_serverinfo |= (SERVERINFO_NTDS|SERVERINFO_CNFPR);
+              else if (nfields > 1
+                       && !ascii_strcasecmp (fields[1], "cnfpr"))
+                *r_serverinfo |= SERVERINFO_CNFPR;
             }
         }
       my_ldap_value_free (vals);
@@ -860,7 +864,6 @@ my_ldap_connect (parsed_uri_t uri, unsigned int generic, LDAP **ldap_connp,
     }
   else if (bindname)
     {
-
       npth_unprotect ();
       /* Older Windows header dont have the const for the last two args.
        * Thus we need to cast to avoid warnings.  */
@@ -2730,7 +2733,7 @@ extract_attributes (LDAPMod ***modlist, int *extract_state,
 gpg_error_t
 ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
 	     void *data, size_t datalen,
-	     void *info, size_t infolen)
+	     void *info_arg, size_t infolen_arg)
 {
   gpg_error_t err = 0;
   int ldap_err;
@@ -2741,9 +2744,8 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
   LDAPMod **addlist = NULL;
   char *data_armored = NULL;
   int extract_state;
-
-  /* The last byte of the info block.  */
-  const char *infoend = (const char *) info + infolen - 1;
+  char *info = info_arg;
+  char *info_buffer = NULL;  /* Used in case we need to copy INFO.  */
 
   /* Enable this code to dump the modlist to /tmp/modlist.txt.  */
 #if 0
@@ -2754,12 +2756,30 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
 #endif
   estream_t dump = NULL;
 
-  /* Elide a warning.  */
-  (void) ctrl;
-
   if (dirmngr_use_tor ())
     return no_ldap_due_to_tor (ctrl);
 
+  /* Make sure that INFO is a proper string. */
+  if (!infolen_arg)
+    {
+      /* Too short - use an empty string.  */
+      info = info_buffer = xtrystrdup ("");
+    }
+  else if (info[infolen_arg-1])
+    {
+      /* No terminating nul.  Take a copy.  */
+      info_buffer = xtrymalloc (infolen_arg + 1);
+      if (!info_buffer)
+        {
+          err = gpg_error_from_syserror ();
+          goto out;
+        }
+      memcpy (info_buffer, info, infolen_arg);
+      info_buffer[infolen_arg] = '\0';
+      info = info_buffer;
+    }
+
+  /* Connect to the LDAP server to get its properties.  */
   err = my_ldap_connect (uri, 0, &ldap_conn, &basedn, NULL, NULL, &serverinfo);
   if (err || !basedn)
     {
@@ -2823,8 +2843,8 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
       if (dump)
 	{
 	  es_fprintf(dump, "data (%zd bytes)\n", datalen);
-	  es_fprintf(dump, "info (%zd bytes): '\n", infolen);
-	  es_fwrite(info, infolen, 1, dump);
+	  es_fprintf(dump, "info (%zd bytes): '\n", strlen (info));
+	  es_fwrite(info, strlen (info), 1, dump);
 	  es_fprintf(dump, "'\n");
 	}
     }
@@ -2851,37 +2871,24 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
 
   /* Assemble the INFO stuff into LDAP attributes */
   extract_state = 0;
-  while (infolen > 0)
+  while (*info)
     {
-      char *temp = NULL;
+      char *newline;
 
-      char *newline = memchr (info, '\n', infolen);
-      if (! newline)
-	/* The last line is not \n terminated!  Make a copy so we can
-	   add a NUL terminator.  */
-	{
-	  temp = xmalloc (infolen + 1);
-	  memcpy (temp, info, infolen);
-	  info = temp;
-	  newline = (char *) info + infolen;
-	}
-
-      *newline = '\0';
+      newline = strchr (info, '\n');
+      if (newline)
+        *newline = '\0';  /* Replace the LF */
 
       extract_attributes (&addlist, &extract_state, info,
                           (serverinfo & SERVERINFO_SCHEMAV2));
 
-      infolen = infolen - ((uintptr_t) newline - (uintptr_t) info + 1);
-      info = newline + 1;
-
-      /* Sanity check.  */
-      if (! temp)
-	log_assert ((char *) info + infolen - 1 == infoend);
+      if (newline)
+        {
+          *newline = '\n';     /* Restore the LF  */
+          info = newline + 1;  /* and forward.    */
+        }
       else
-	{
-	  log_assert (infolen == -1);
-	  xfree (temp);
-	}
+        info += strlen (info); /* Forward to the Nul.  */
     }
 
   modlist_add (&addlist, "objectClass", "pgpKeyInfo");
@@ -2918,7 +2925,7 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
     char **attrval;
     char *dn;
 
-    if ((serverinfo & SERVERINFO_NTDS))
+    if ((serverinfo & SERVERINFO_CNFPR))
       {
         /* The modern way using a CN RDN with the fingerprint.  This
          * has the advantage that we won't have duplicate 64 bit
@@ -2983,7 +2990,117 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
   xfree (addlist);
 
   xfree (data_armored);
+  xfree (info_buffer);
+  return err;
+}
 
+
+/* Delete the keys given by the list of fingerprints in FPRLIST from
+ * the keyserver identified by URI.  The function stops at the first
+ * error encountered. */
+gpg_error_t
+ks_ldap_del (ctrl_t ctrl, parsed_uri_t uri, strlist_t fprlist)
+{
+  gpg_error_t err = 0;
+  int ldap_err;
+  unsigned int serverinfo;
+  LDAP *ldap_conn = NULL;
+  char *basedn = NULL;
+  char *dn = NULL;
+  strlist_t fpr;
+  unsigned int count = 0;
+  unsigned int totalcount = 0;
+
+  if (dirmngr_use_tor ())
+    {
+      return no_ldap_due_to_tor (ctrl);
+    }
+
+  for (fpr = fprlist; fpr; fpr = fpr->next)
+    totalcount++;
+
+  err = my_ldap_connect (uri, 0, &ldap_conn, &basedn, NULL, NULL, &serverinfo);
+  if (err || !basedn)
+    {
+      if(opt.verbose)
+        log_info ("%s: connecting to server failed\n", __func__);
+      if (!err)
+	err = gpg_error (GPG_ERR_GENERAL);  /* (no baseDN) */
+      goto leave;
+    }
+
+  if (!(serverinfo & SERVERINFO_REALLDAP))
+    {
+      if(opt.verbose)
+        log_info ("%s: The PGP.com keyserver is not supported\n", __func__);
+      err = gpg_error (GPG_ERR_NOT_SUPPORTED);
+      goto leave;
+    }
+
+  if (!(serverinfo & SERVERINFO_SCHEMAV2))
+    {
+      if(opt.verbose)
+        log_info ("%s: The keyserver does not support the v2 schema\n",
+                  __func__);
+      err = gpg_error (GPG_ERR_NOT_SUPPORTED);
+      goto leave;
+    }
+
+  if (opt.verbose)
+    log_info ("%s: Using DN: %s,%s\n", __func__,
+              (serverinfo & SERVERINFO_CNFPR)? "CN=<fingerprint>"
+              /* */                          : "pgpCertID=<keyid>",
+              basedn);
+  for (fpr = fprlist; fpr; fpr = fpr->next)
+    {
+      if ((serverinfo & SERVERINFO_CNFPR))
+        {
+          xfree (dn);
+          dn = xtryasprintf ("CN=%s,%s", fpr->d, basedn);
+        }
+      else
+        {
+          unsigned int off;
+
+          /* Simle method to get the keyID.  Note that a v5 key
+           * (len>40) has the keyid at the left.  If the length is
+           * less than 17 we assume a keyid has been given.  */
+          off = strlen (fpr->d);
+          if (off <= 40 && off > 16)
+            off = off - 16;
+          else
+            off = 0;
+
+          xfree (dn);
+          dn = xtryasprintf ("pgpCertID=%.16s,%s", fpr->d+off, basedn);
+        }
+
+      npth_unprotect ();
+      ldap_err = ldap_delete_ext_s (ldap_conn, dn, NULL, NULL);
+      npth_protect ();
+      if (ldap_err == LDAP_SUCCESS)
+        {
+          if (opt.verbose)
+            log_info ("%s: key %s deleted\n", __func__, fpr->d);
+          count++;
+        }
+      else
+        {
+          log_error ("%s: error deleting key %s: %s\n",
+                     __func__, fpr->d, ldap_err2string (ldap_err));
+          err = ldap_err_to_gpg_err (ldap_err);
+          break;  /* Stop at the first failed deletion.  */
+        }
+    }
+  log_info ("%s: number of keys deleted: %u of %u\n",
+            __func__, count, totalcount);
+
+
+ leave:
+  if (ldap_conn)
+    ldap_unbind (ldap_conn);
+  xfree (dn);
+  xfree (basedn);
   return err;
 }
 
