@@ -61,7 +61,7 @@ struct symlist_item
 {
   struct symlist_item *next;
   int cipher_algo;
-  int cfb_mode;
+  enum gcry_cipher_modes cipher_mode;
   int other_error;
 };
 
@@ -495,7 +495,11 @@ proc_symkey_enc (CTX c, PACKET *pkt)
     if (enc)
       {
         symitem->cipher_algo = enc->cipher_algo;
-        symitem->cfb_mode = !enc->aead_algo;
+        symitem->cipher_mode = !enc->aead_algo;
+        symitem->cipher_mode
+          = (enc->aead_algo == AEAD_ALGO_NONE? GCRY_CIPHER_MODE_CFB :
+             enc->aead_algo == AEAD_ALGO_OCB?  GCRY_CIPHER_MODE_OCB :
+             GCRY_CIPHER_MODE_NONE);
       }
     else
       symitem->other_error = 1;
@@ -675,6 +679,8 @@ proc_encrypted (CTX c, PACKET *pkt)
   int result = 0;
   int early_plaintext = literals_seen;
   unsigned int compliance_de_vs = 0;
+  enum gcry_cipher_modes ciphermode;
+  int unknown_ciphermode;
 
   if (pkt->pkttype == PKT_ENCRYPTED_AEAD)
     c->seen_pkt_encrypted_aead = 1;
@@ -764,14 +770,26 @@ proc_encrypted (CTX c, PACKET *pkt)
   else if (!c->dek)
     result = GPG_ERR_NO_SECKEY;
 
+  /* We need to know the ciphermode for gnupg_cipher_is_compliant.  */
+  unknown_ciphermode = 0;
+  if (pkt->pkt.encrypted->aead_algo)
+    {
+      unsigned int dummy;
+      if (openpgp_aead_algo_info (pkt->pkt.encrypted->aead_algo,
+                                  &ciphermode, &dummy))
+        unknown_ciphermode = 1;  /* error -> unknown mode */
+    }
+  else
+    ciphermode = GCRY_CIPHER_MODE_CFB;
+
   /* Compute compliance with CO_DE_VS.  */
   if (!result && (is_status_enabled () || opt.flags.require_compliance)
       /* Overriding session key voids compliance.  */
       && !opt.override_session_key
       /* Check symmetric cipher.  */
       && gnupg_gcrypt_is_compliant (CO_DE_VS)
-      && gnupg_cipher_is_compliant (CO_DE_VS, c->dek->algo,
-                                    GCRY_CIPHER_MODE_CFB))
+      && !unknown_ciphermode
+      && gnupg_cipher_is_compliant (CO_DE_VS, c->dek->algo, ciphermode))
     {
       struct kidlist_item *i;
       struct symlist_item *si;
@@ -786,9 +804,9 @@ proc_encrypted (CTX c, PACKET *pkt)
        * key can be sneaked in.  */
       for (si = c->symenc_list; si && compliant; si = si->next)
         {
-          if (!si->cfb_mode
+          if (si->cipher_mode == GCRY_CIPHER_MODE_NONE
               || !gnupg_cipher_is_compliant (CO_DE_VS, si->cipher_algo,
-                                             GCRY_CIPHER_MODE_CFB))
+                                             si->cipher_mode))
             compliant = 0;
         }
 
