@@ -71,6 +71,7 @@ static int lock_all (KEYDB_HANDLE hd);
 static void unlock_all (KEYDB_HANDLE hd);
 
 
+
 static void
 try_make_homedir (const char *fname)
 {
@@ -340,10 +341,14 @@ keydb_add_resource (ctrl_t ctrl, const char *url, int force, int *auto_created)
                 KEYBOX_HANDLE kbxhd = keybox_new_x509 (token, 0);
 
                 if (kbxhd)
-                  {
-                    keybox_compress (kbxhd);
-                    keybox_release (kbxhd);
-                  }
+                  keybox_compress (kbxhd);
+                keybox_release (kbxhd);
+                /* The release of the lock is okay here because we
+                 * known that this process has currently no other
+                 * users of the lock.  In general we should change the
+                 * simple already-locked optimization to use reference
+                 * counting.  But that will introduce new sources for
+                 * error locking errors. */
                 dotlock_release (all_resources[used_resources].lockhandle);
               }
 
@@ -438,7 +443,6 @@ keydb_release (KEYDB_HANDLE hd)
   assert (active_handles > 0);
   active_handles--;
 
-  unlock_all (hd);
   for (i=0; i < hd->used; i++)
     {
       switch (hd->active[i].type)
@@ -450,8 +454,9 @@ keydb_release (KEYDB_HANDLE hd)
           break;
         }
     }
+  unlock_all (hd);
 
-    xfree (hd);
+  xfree (hd);
 }
 
 
@@ -749,8 +754,8 @@ keydb_set_flags (KEYDB_HANDLE hd, int which, int idx, unsigned int value)
 /*
  * Insert a new Certificate into one of the resources.
  */
-int
-keydb_insert_cert (KEYDB_HANDLE hd, ksba_cert_t cert)
+static int
+do_insert_cert (KEYDB_HANDLE hd, ksba_cert_t cert)
 {
   int rc = -1;
   int idx;
@@ -784,16 +789,16 @@ keydb_insert_cert (KEYDB_HANDLE hd, ksba_cert_t cert)
       break;
     }
 
-  unlock_all (hd);
   return rc;
 }
 
 
 /*
- * The current keyblock or cert will be deleted.
+ * The current keyblock or cert will be deleted.  Note that this needs
+ * to be called file the file lock held.
  */
 int
-keydb_delete (KEYDB_HANDLE hd, int unlock)
+keydb_delete (KEYDB_HANDLE hd)
 {
   int rc = -1;
 
@@ -819,8 +824,6 @@ keydb_delete (KEYDB_HANDLE hd, int unlock)
       break;
     }
 
-  if (unlock)
-    unlock_all (hd);
   return rc;
 }
 
@@ -941,9 +944,10 @@ keydb_search (ctrl_t ctrl, KEYDB_HANDLE hd,
   rc = lock_all (hd);
   if (rc)
     return rc;
-  rc = -1;
+  rc = gpg_error (GPG_ERR_EOF);
 
-  while (rc == -1 && hd->current >= 0 && hd->current < hd->used)
+  while (gpg_err_code (rc) == GPG_ERR_EOF
+         && hd->current >= 0 && hd->current < hd->used)
     {
       switch (hd->active[hd->current].type)
         {
@@ -954,9 +958,11 @@ keydb_search (ctrl_t ctrl, KEYDB_HANDLE hd,
           rc = keybox_search (hd->active[hd->current].u.kr, desc, ndesc,
                               KEYBOX_BLOBTYPE_X509,
                               NULL, &skipped);
+          if (rc == -1) /* Map legacy code.  */
+            rc = gpg_error (GPG_ERR_EOF);
           break;
         }
-      if (rc == -1 || gpg_err_code (rc) == GPG_ERR_EOF)
+      if (gpg_err_code (rc) == GPG_ERR_EOF)
         { /* EOF -> switch to next resource */
           hd->current++;
         }
@@ -964,6 +970,10 @@ keydb_search (ctrl_t ctrl, KEYDB_HANDLE hd,
         hd->found = hd->current;
     }
 
+  /* The NOTHING_FOUND error is triggered by a NEXT command.  */
+  if (gpg_err_code (rc) == GPG_ERR_EOF
+      || gpg_err_code (rc) == GPG_ERR_NOTHING_FOUND)
+    rc = gpg_error (GPG_ERR_NOT_FOUND);
   return rc;
 }
 
@@ -1097,12 +1107,14 @@ keydb_store_cert (ctrl_t ctrl, ksba_cert_t cert, int ephemeral, int *existed)
   keydb_set_ephemeral (kh, 1);
 
   keydb_close_all_files ();
+  unlock_all (kh);
+  gnupg_usleep (20);  /* Give other processes a chance to take the lock.  */
   rc = lock_all (kh);
   if (rc)
     return rc;
 
   rc = keydb_search_fpr (ctrl, kh, fpr);
-  if (rc != -1)
+  if (gpg_err_code (rc) != GPG_ERR_NOT_FOUND)
     {
       keydb_release (kh);
       if (!rc)
@@ -1141,7 +1153,7 @@ keydb_store_cert (ctrl_t ctrl, ksba_cert_t cert, int ephemeral, int *existed)
       return rc;
     }
 
-  rc = keydb_insert_cert (kh, cert);
+  rc = do_insert_cert (kh, cert);
   if (rc)
     {
       log_error (_("error storing certificate: %s\n"), gpg_strerror (rc));
@@ -1194,9 +1206,7 @@ keydb_set_cert_flags (ctrl_t ctrl, ksba_cert_t cert, int ephemeral,
   err = keydb_search_fpr (ctrl, kh, fpr);
   if (err)
     {
-      if (err == -1)
-        err = gpg_error (GPG_ERR_NOT_FOUND);
-      else
+      if (gpg_err_code (err) != gpg_error (GPG_ERR_NOT_FOUND))
         log_error (_("problem re-searching certificate: %s\n"),
                    gpg_strerror (err));
       keydb_release (kh);
@@ -1282,6 +1292,8 @@ keydb_clear_some_cert_flags (ctrl_t ctrl, strlist_t names)
     }
 
   keydb_close_all_files ();
+  unlock_all (hd);
+  gnupg_usleep (20);  /* Give other processes a chance to take the lock.  */
   err = lock_all (hd);
   if (err)
     {
@@ -1313,7 +1325,7 @@ keydb_clear_some_cert_flags (ctrl_t ctrl, strlist_t names)
             }
         }
     }
-  if (rc && rc != -1)
+  if (rc && gpg_err_code (rc) != GPG_ERR_NOT_FOUND)
     log_error ("%s failed: %s\n", __func__, gpg_strerror (rc));
 
  leave:
